!==================================================================================================================================
! LVDSMC_BGK_1D.f90
!==================================================================================================================================

!   Low-variance direct simulation Monte Carlo (LVDSMC) solver.
!   written by: Gregg A. Radtke

!----------------------------------------------------------------------------------------------------------------------------------
! DISCLAIMER: This computational code is provided as is without any guarantees or warranty.  Although the author has attempted to
!   find and correct any bugs, the author is not responsible for any damage or losses of any kind caused by the use or misuse of
!   the code.  The author is under no obligation to provide support, service, corrections, or upgrades.
!----------------------------------------------------------------------------------------------------------------------------------

!----------------------------------------------------------------------------------------------------------------------------------
! REQUIRES:
!----------------------------------------------------------------------------------------------------------------------------------

!   (1) modern FORTRAN compiler
!   (2) input file ("input.inp"), in the format described below
!   (3) ratio-of-uniforms bounding factors file ("bounds.inp"), in the format described below

!   The program can be compiled with any modern FORTRAN compiler.  A free FORTRAN compiler (G95) is available from 
!   http://www.g95.org.  This code has been successfully compiled and run on G95 version 0.92.


!----------------------------------------------------------------------------------------------------------------------------------
! FEATURES:
!----------------------------------------------------------------------------------------------------------------------------------

!   BGK collision model
!   One-dimensional, rectangular geometry
!   Boundary conditions: Maxwell's accommodation model
!   Streamwise pressure and temperature gradients
!   Dimensionless variables


!----------------------------------------------------------------------------------------------------------------------------------
! VERSION HISTORY:
!----------------------------------------------------------------------------------------------------------------------------------

!   1.0      Jan 28, 2011

!----------------------------------------------------------------------------------------------------------------------------------
! DIMENSIONLESS VARIABLES:
!----------------------------------------------------------------------------------------------------------------------------------

!   dimensional parameters
!   ----------------------
!   f0       reference equilibrium distribution [kg s^3/m^6]
!   rho0     reference density [kg/m^3]
!   T0       reference temperature [K]
!   c0       most probable molecular speed at reference conditions [m/s]
!   tau0     relaxation time at reference conditions [s]
!   m        molecular mass [kg]
!   L0       system length scale [m]
!   R        gas constant ( = kB/m, where kB is the Boltzmann constant) [J/(kg K)]
!   c        molecular velocity (vector) [m/s]
!   x        spatial position (0 <= x <= L0)
!   rho      mass density [kg/m^3]
!   u        velocity [m/s]
!   P        pressure tensor [N/m]
!   T        temperature [K]
!   q        heat flux [W/m^2]

!   reference equilibrium distribution (also the initial condition)
!   ---------------------------------------------------------------
!   f0 = rho0 / ( pi^(3/2) c0^3) exp ( - || c ||^2 / c0^2 )
!     where c0 = sqrt ( 2 R T0 )

!   Knudsen number:
!   ---------------
!   Kn = [ 2 / sqrt(pi) ] ( c0 tau0 / L0 )
!     where 1 / l0 = sqrt(2) pi ( rho0 / m ) d0^2

!   dimensionless spatial position, time, particle velocity
!   -------------------------------------------------------
!   x* = x / L0
!   t* = c0 t / L0
!   c* = c / c0

!   dimensionless hydrodynamic variables
!   ------------------------------------
!   rho* = ( rho / rho0 - 1 ) / epsilon
!   u*   = ( u   / c0 ) / epsilon
!   P*   = [ P   / ( rho0 R T0 )   - I ] / epsilon     (where I is the 3x3 identity tensor)
!   T*   = ( T   / T0   - 1 ) / epsilon
!   q*   = [ q   / ( c0 rho0 R T0 ) ] / epsilon


!----------------------------------------------------------------------------------------------------------------------------------
! INPUT FILE:     ( file name = "input" )
!----------------------------------------------------------------------------------------------------------------------------------

!   input file format
!   -----------------
!   Nt, dt, Nout, Nss
!   Nx, Lx, Xi, Nens, disp
!   Kn, epsilon, kappa_P, kappa_T
!   type_b1, alpha_b1, uy*_b1, uz*_b1, T*_b1
!   type_b2, alpha_b2, uy*_b2, uz*_b2, T*_b2

!   input parameters
!   ----------------
!   Nt          Total number of time steps
!   dt          Time step in percentage of cell traversal, i.e.: dt = c0 dt* / dx*
!   Nout        Number of time steps between outputing properties
!   Nss         Number of time steps until steady state conditions are reached
!   Nx          Number of computational cells in the x direction (regular grid)
!   Lx          Simulation length in dimensionless form: Lx = Lx* / L0, either 1.0 or 0.5
!                 for most simulations, choose Lx = 1.0
!                 when one of the walls is a symmetry condition (BC type = 0), choose Lx = 0.5
!   Xi          Parameter that is approximately equal to the number of particles per cell (~10 for typical problems)
!   Nens        Number of independent statistical ensembles
!   disp        Option for the type of simulation progress information displayed on screen:
!                 (1) display simulation information and endpoint only
!                 (2) display after each statistical ensemble is finished
!                 (3) display after each recorded data point (every Nout steps)
!   Kn          Knudsen number based on the system length scale: Kn = l0 / L0
!   epsilon  Departure from equilibrium, which scales all the boundary conditions and streamwise pressure/temperature gradient
!              driving forces.  Note: it is important that the fundamental (largest) nonhomogeniety in the problem (i.e. stream-
!              wise gradients: kappa_P, kappa_T; or boundary conditions: uy*_, uz*_, T*_; discussed below) be chosen to be unity
!              in absolute value.  Thus the departure from equilibrium is to be independently controlled via epsilon without 
!              affecting the overall configuration of the problem.  Following this procedure rigorously is also important to ensure
!              that the parameter Xi (above) is the same order of magnitude of particles in simulation.
!   kappa_P  Dimensionless pressure gradient in the streamwise (z) direction: Pz = - ( L0 / P0 ) ( dP / dz ) / epsilon
!              this is the driving force for Poiseulle flow
!   kappa_T  Dimensionless temperature gradient in the streamwise (z) direction: Tz = ( L0 / T0 ) ( dT / dz ) / epsilon
!              this is the driving force for thermal creep
!   b1, b2   Subscripts refering to the two boundary condtions: b1 corresponds to x = 0 and b2 corresponds to x = Lx
!   type_    Boundary condition type:
!              (0) specular
!              (1) diffusely-reflecting
!              (2) Maxwell's accomodation model
!   alpha_   Accomodation coefficient: the probability ( 0.0 < alpha < 1.0 ) of diffuse reflection
!   uy*_     Dimensionless boundary velocity in the y-direction (see above definition)
!   uz*_     Dimensionless boundary velocity in the z-direction (see above definition)
!   T*_      Dimensionless boundary temperature (see above definition)


!----------------------------------------------------------------------------------------------------------------------------------
! SCREEN OUTPUT:
!----------------------------------------------------------------------------------------------------------------------------------

!   When the compiled program is run, various input data is displayed on screen, consisting of simulation parameters and boundary
!   conditions.  Most of these parameters should be clear from the discussion of input file parameters above.  Two additional
!   parameters are calculated and displayed: advective time scales (t* = c0 t / L0) for both the total simulation time, as well as
!   for the time at which the simulation will assume that steady state conditions are reached and begin steady state averaging.
!   For simulations with Kn > 0.1, reasonable results for steady state calculations are obtained when the steady state advective
!   time scale (tSS*) is greater than 50.  For Kn < 0.1, steady state conditions are determined by the diffusive (rather than
!   advective) time scale, and reasonable results are obtained for tSS* Kn > 5.

!   During operation, simulation progress is output according to the input parameter "disp" discussed above.  Each progress update
!   is indicated with an appropriate time stamp (processor time), and also indicates the current number of deviational particles in
!   simulation.  For typical simulations in the transition regime (0.1 < Kn < 10) with appropriately chosen epsilon and properly
!   specified boundary condtions and streamwise gradients, this value should be of the same order of magnitide as the product of
!   the Xi parameter and the number of cells: Nx * Xi.


!----------------------------------------------------------------------------------------------------------------------------------
! OUTPUT FILES:
!----------------------------------------------------------------------------------------------------------------------------------

!   spatial coordinates ( file name = "x.dat" )
!   -------------------------------------------
!   x_1
!   x_2
!   ...
!   x_Nx

!   cell-based hydrodynamic properties ( file name = "output.dat" )
!   ---------------------------------------------------------------
!   rho  ux  uy  uz  Pxx  Pxy  Pyy  Pyz  Pzz  Pzx  T  qx  qy  qz

!     if Nt <= Nss in the input file, there will be Nt/Nout sets of Nx values corresponding to each transient time step
!     if Nt > Nss in the input file, there will be a single set of Nx values for the steady state averaged properties


!----------------------------------------------------------------------------------------------------------------------------------
! RATIO OF UNIFORMS BOUNDING FACTORS:     ( file name = "bounds" )
!----------------------------------------------------------------------------------------------------------------------------------

!   The bounding factors are used for sampling the particles created at the boundaries due to advection.  These factors are
!   dynamically tested (when r_Y > 0) and updated when they are exceeded, and each instance will be accompanied by an on-screen
!   warning.  If the simulation finishes normally, the "bounds" file will be rewritten with the updated values.  If a significant
!   number of warning messages are encountered in a simulation, the results may be in error, and the simulation should be rerun.
!   Note that the bounds used for sampling the BGK collision operator may require significantly larger corrections, even in the
!   linearized case.  This is reflected in the list of typical values (below).

!   bounds file format
!   ------------------
!   Y_a_adv,  Y_bn_adv, Y_bt_adv, r_adv
!   Y_a_coll, Y_b_coll, r_coll

!   typical values
!   --------------
!   1.0, 1.0, 1.0, 0.01
!   1.0, 1.0, 0.05

!   bounding factors
!   ----------------
!   Y_a_adv     Bounding factor for the transformed dependent variable for advection routine
!   Y_bn_adv    Bounding factor for the transformed particle velocity in the normal (x) direction for the advection routine
!   Y_bt_adv    Bounding factor for the transformed particle velocity in the tangential (y,z) directions for the advection routine
!   r_adv       Percentage of oversampling (0 <= r_adv << 1): the bounds are further modified by a small percentage in order to
!                 test the validity of the bounding factors.
!   r_adv       Percentage of oversampling (0 <= r_Y << 1): the bounds are further modified by a small percentage in order to test
!                 the validity of the bounding factors.  For linearized conditions (epsillon << 1) this can be set to zero if all
!                 bounding factors (above) are set to unity.  Otherwise, set to a small percentage.
!   Y_a_coll    Bounding factor for the transformed dependent variable for collision routine
!   Y_b_coll    Bounding factor for the transformed particle velocity for the collision routine
!   r_coll      Percentage of oversampling (0 <= r_coll << 1): similar to r_adv.


!----------------------------------------------------------------------------------------------------------------------------------
! NUMERICAL PARAMETERS USED IN THE CODE:
!----------------------------------------------------------------------------------------------------------------------------------

!   There are several computational parameters hardwired into the code listing, which can be found within the module "numerical
!   constants".  While the default values should be appropriate for most platforms and typical simulations, they can be modified
!   (if desired) to accomodate special simulations.  When these values are exceeded, the user is alerted (by an appropriate error
!   message on the screen), and the simulation stops.

!   computational parameters
!   ------------------------
!   Nmax      Maximum total number of particles in simulation
!   Ncmax     Maximum number of particles contained within any computational cell
!   Nbmax     Maximum number of particles to strike a boundary within a time step


!==================================================================================================================================
! MODULES
!==================================================================================================================================

! NUMERICAL CONSTANTS
  module numerical_constants
    real(8), parameter :: pi = 3.141592653589793d0
    real(8), parameter :: srpi = 1.772453850905516d0
    real(8), parameter :: eps1 = 1.d-14

!   computational parameters
    integer, parameter :: Nmax = 1000000
    integer, parameter :: Ncmax = 10000
    integer, parameter :: Nbmax = 10000

  end module numerical_constants

!----------------------------------------------------------------------------------------------------------------------------------
! RATIO OF UNIFORMS SAMPLING BOUND CONSTANTS
  module sampling_bound_constants

!   bounds for fMB
    real(8), parameter :: a0 = 1.d0
    real(8), parameter :: b0 = 0.811173616822835d0

!   bounds for c_n fMB
    real(8), parameter :: an   = 0.428881942480353d0
    real(8), parameter :: bn_n = 1.344250845932327d0
    real(8), parameter :: bn_s = 0.347897716471792d0

!   bounds for c_n c_n fMB
    real(8), parameter :: ann   = 0.367879441171442d0
    real(8), parameter :: bnn_n = 2.422185882560844d0
    real(8), parameter :: bnn_s = 0.298414096849802d0

!   bounds for c_n c_s fMB
    real(8), parameter :: ans   = 0.183939720585721d0
    real(8), parameter :: bns_n = 0.576524913984314d0
    real(8), parameter :: bns_t = 0.149207048424901d0

!   bounds for c2 fMB
    real(8), parameter :: a2 = 0.367879441171442d0
    real(8), parameter :: b2 = 2.422185882560844d0

!   bounds for c2 c_n fMB
    real(8), parameter :: a2n   = 0.409916278941860d0
    real(8), parameter :: b2n_n = 4.688803555515950d0
    real(8), parameter :: b2n_s = 1.213481884638469d0

  end module sampling_bound_constants

!----------------------------------------------------------------------------------------------------------------------------------
! GLOBAL DATA TYPES
  module types

!   data type for boundary conditions
    type TBC
      integer, dimension(2) :: type
      integer, dimension(2,3) :: norm
      real(8), dimension(2) :: p, c, n, a, Ns, na, ca, x
      real(8), dimension(2,3) :: u, b
    end type TBC

!   data type for numerical parameters for bounding sampled distributions
    type Tbounds
      real(8) :: Ya_adv, Ya_coll, Yb_coll, r_adv, r_coll
      real(8), dimension(2) :: Yb_adv
    end type Tbounds

!   data type for boundary element based particle cross referencing
    type Tbref
      integer, dimension(0:1,2) :: N
      integer, dimension(:,:,:), allocatable :: ref
    end type Tbref

!   data type for cell based particle cross referencing
    type Tcref
      logical, dimension(:,:), allocatable :: G
      integer :: DS
      integer, dimension(0:1) :: NG
      integer, dimension(:), allocatable :: S
      integer, dimension(:,:), allocatable :: N, idx, Gidx
      integer, dimension(:,:,:), allocatable :: ref, Glist
      real(8) :: Jmax, Jsum
      real(8), dimension(:), allocatable :: x, C, a, F, na, ca, J
      real(8), dimension(:,:), allocatable :: U, b, ua
    end type Tcref

!   data type for Maxwell-Boltzmann distribution
    type TfMB
      real(8) :: n, c
      real(8), dimension(3) :: u
    end type TfMB

!   data type for Maxwell-Boltzmann distribution
    type Tfloc
      real(8), dimension(:), allocatable :: n, c
      real(8), dimension(:,:), allocatable :: u
    end type Tfloc

!   data type for particle data
    type Tpart
      logical, dimension(:,:), allocatable :: del
      integer, dimension(0:1) :: N
      integer, dimension(:,:), allocatable :: cell
      real(8), dimension(:,:), allocatable :: x
      real(8), dimension(:,:,:), allocatable :: c
    end type Tpart

!   data type for property data
    type Tprop
      real(8), dimension(:), allocatable :: n, T
      real(8), dimension(:,:), allocatable :: u, p, q
    end type Tprop

!   data type for simulation parameters
    type Tsim
      logical :: force
      integer :: Nt, Nout, Nss, Nens, disp, Ncell
      real(8) :: dt, Xi, Kn, eps, af, Nf, Pz, Tz, dx, Lx
      real(8), dimension(3) :: bf
    end type Tsim

!   data type for time step
    type Ttime
      real(8) :: t, tc, dtc
    end type Ttime
  end module types


!==================================================================================================================================
! MAIN PROGRAM
!==================================================================================================================================

  program LVDSMC
    use numerical_constants
    use types
    implicit none

!   variable declarations
    integer :: step, ens
    real(8), dimension(:,:,:), allocatable :: data
    type(Tbref) :: bref
    type(TBC) :: BC
    type(Tcref) :: cref
    type(TfMB) :: f0
    type(Tfloc) :: floc
    type(Tbounds) :: bounds
    type(Tpart) :: part
    type(Tprop) :: trans, steady
    type(Tsim) :: sim
    type(Ttime) :: time

!   read input data
    call input_data (BC, bref, cref, data, f0, floc, part, trans, steady, sim, time)

!   perform simulation over all ensembles
    do ens=1,sim%Nens

!     initialize state
      call initialize_sim (cref, f0, bounds, part, sim, time)

!     time steps
      do step=1,sim%Nt

!       perform time steps
        time%t = step*sim%dt
        call time_step (BC, bref, f0, floc, bounds, part, trans, steady, sim, step, time)

!       update properties
        call properties (f0, part, trans, sim)
        if (sim%Nt <= sim%Nss) then

!         record transient properties
          if (mod(step,sim%Nout)==0) call record_prop_trans (step/sim%Nout, sim%Ncell, data, trans)
        end if
        if ( (sim%disp>2) .AND. (mod(step,sim%Nout)==0) ) call display_rec (step, sim%Nt, sum(part%N))
        if (step >= sim%Nss) then

!         record data for steady state properties
          call record_prop_steady (trans, steady, sim)
        end if
      end do

!     display time step after each ensemble is finished
      if (sim%disp>1) call display_ens (ens, sim%Nens, sum(part%N), sim%disp)
    end do

!   output ensemble averaged data
    call display_out (sum(part%N))
    if (sim%Nt>sim%Nss) then
      call output_prop_steady (steady, sim)
    else
      call output_prop_trans (data, sim)
    end if

!   end simulation
    call end_sim (sum(part%N), bounds)

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ALLOCATE DATA STORAGE ARRAYS
    subroutine allocate_arrays (bref, cref, data, floc, part, trans, steady, sim)
      implicit none
      integer :: Ncell, Ndt, i, j
      real(8), dimension(:,:,:), allocatable :: data
      type(Tbref) :: bref
      type(Tcref) :: cref
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tprop) :: trans, steady
      type(Tsim) :: sim

!     number of cells
      Ncell = sim%Ncell

!     allocate arrays
      allocate ( bref%ref(0:1,2,Nbmax) )
      allocate ( cref%N(0:1,Ncell), cref%S(Ncell), cref%idx(0:1,Nmax), cref%ref(0:1,Ncell,Ncmax), cref%x(Ncell), cref%C(Ncell),   &
                   cref%U(Ncell,3), cref%G(0:1,Nmax), cref%Gidx(0:1,Nmax), cref%Glist(0:1,1,Nmax), cref%a(Ncell), cref%b(Ncell,3),&
                                                   cref%na(Ncell), cref%ua(Ncell,3), cref%ca(Ncell), cref%J(Ncell), cref%F(Ncell) )
      allocate ( floc%n(Ncell), floc%u(Ncell,3), floc%c(Ncell) )
      allocate ( part%del(0:1,Nmax), part%x(0:1,Nmax), part%c(0:1,Nmax,3), part%cell(0:1,Nmax))
      allocate ( trans%n(Ncell), trans%u(Ncell,3), trans%T(Ncell), trans%p(Ncell,6), trans%q(Ncell,3) )
      allocate ( steady%n(Ncell), steady%u(Ncell,3), steady%T(Ncell), steady%p(Ncell,6), steady%q(Ncell,3) )

!     data structure for ensemble averaged transient data
      if ( sim%Nt <= sim%Nss ) then
        Ndt = sim%Nt / sim%Nout
        allocate ( data(Ncell,Ndt,14) )
        do i = 1, Ncell
          do j = 1, Ndt
            data (i,j,:) = (/0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0/)
          end do
        end do
      end if

!     zero out steady properties arrays
      do i = 1, Ncell
        steady%n(i) = 0.d0
        steady%u(i,:) = (/0.d0,0.d0,0.d0/)
        steady%T(i) = 0.d0
        steady%p(i,:) = (/0.d0,0.d0,0.d0,0.d0,0.d0,0.d0/)
        steady%q(i,:) = (/0.d0,0.d0,0.d0/)
      end do

    end subroutine allocate_arrays
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY END TIME STAMP
    subroutine display_end (Np)
      implicit none
      integer :: Np
      real :: tcpu

      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a6)') '[',tcpu,Np,']: end'
      write (0, '(a48)') '================================================'
      flush (0)

    end subroutine display_end
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY END TIME STAMP
    subroutine display_ens (ens, Nens, Np, disp)
      implicit none
      integer :: ens, Nens, Np, disp
      real :: tcpu

      call cpu_time(tcpu)
      if (disp>2) write (0, '(a48)') '------------------------------------------------'
      write (0, '(a1,es9.3,i9,a11,i7,a3,i7)') '[',tcpu,Np,']: ensemble',ens,' of',Nens
      if (disp>2) write (0, '(a48)') '------------------------------------------------'
      flush (0)

    end subroutine display_ens
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY INITIAL TIME STAMP
    subroutine display_init (Np)
      implicit none
      integer :: Np
      real :: tcpu

      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a24)') '[',tcpu,Np,']: initial condition set'
      write (0, '(a48)') '------------------------------------------------'
      flush (0)

    end subroutine display_init
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY TIME STAMP FOR WRITING THE OUTPUT FILE
    subroutine display_out (Np)
      implicit none
      integer :: Np
      real :: tcpu

      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a22)') '[',tcpu,Np,']: writing output file'
      flush (0)

    end subroutine display_out
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY INITIAL SIMULATION PARAMETERS
    subroutine display_sim (BC, sim)
      implicit none
      real :: tcpu
      type(TBC) :: BC
      type(Tsim) :: sim

!     initialize cpu time
      call cpu_time(tcpu)

!     header
      write (0, *)
      write (0, *)
      write (0, '(a48)') '================================================'
      write (0, '(a29)') 'LVDSMC_BGK_1D.f90 version 1.0'
      write (0, '(a42)') 'low-noise 1D Boltzmann equation simulation'
      write (0, '(a37)') 'Bhatnagar-Gross-Krook collision model'
      write (0, '(a26)') '  written by: Gregg Radtke'
      write (0, '(a48)') '------------------------------------------------'
      write (0, *)

!     general simulation parameters
      write (0, '(a22)') 'simulation parameters:'
      write (0, '(a48)') '================================================'
      write (0, '(a28,es10.3)') 'Knudsen number:             ', sim%Kn
      write (0, '(a28,es10.3)') 'Departure from equilibrium: ', sim%eps
      write (0, '(a28,es10.3)') 'Pressure gradient (z):      ', sim%Pz
      write (0, '(a28,es10.3)') 'Temperature gradient (z):   ', sim%Tz
      write (0, '(a28,es10.3)') 'Simulation domain size:     ', sim%Lx
      write (0, '(a28,es10.3)') 'Approx. particles per cell: ', sim%Xi
      write (0, '(a28,es10.3)') 'Advective time (final):     ', sim%dt * sim%Nt
      write (0, '(a28,es10.3)') 'Advective time (steady):    ', sim%dt * sim%Nss
      write (0, '(a29,i8)') 'Total number of cells:       ', sim%Ncell

!     averaging method
      if ( sim%Nt > sim%Nss ) then
        write (0, '(a42)') 'Steady state averaging (steady simulation)'
      else
        write (0, '(a41)') 'Ensemble averaging (transient simulation)'
      end if
      write (0, '(a29,i8)') 'Number of ensembles:         ', sim%Nens
      write (0, '(a48)') '------------------------------------------------'
      write (0, *)


!     boundary condtitions
      write (0, '(a20)') 'boundary conditions:'
      write (0, '(a48)') '================================================'

!     left boundary
      select case ( BC%type(1) )
      case (0)
        write (0, '(a48)') 'BC type (x = 0):           specular reflection  '
      case (1)
        write (0, '(a48)') 'BC type (x = 0):            diffuse reflection  '
      case (2)
        write (0, '(a48)') 'BC type (x = 0):   Maxwell''s accommodation model'
        if ( (BC%p(1)<0.d0) .OR. (BC%p(1)>1.d0) ) stop 'ERROR: inappropriate accommodation coefficient specified for x = 0'
        write (0, '(a28,es10.3)') 'Accomodation coefficient:   ', BC%p(1)
      case default
        write (0, '(a48)') '------------------------------------------------'
        stop 'ERROR: incorrect boundary condition type specified'
      end select
      write (0, '(a28,es10.3)') 'y-velocity:                 ', BC%u(1,2)
      write (0, '(a28,es10.3)') 'z-velocity:                 ', BC%u(1,3)
      write (0, '(a28,es10.3)') 'temperature:                ', BC%c(1)
      write (0, '(a48)') '------------------------------------------------'

!     right boundary
      select case ( BC%type(2) )
      case (0)
        write (0, '(a48)') 'BC type (x = Lx):          specular reflection  '
      case (1)
        write (0, '(a48)') 'BC type (x = Lx):           diffuse reflection  '
      case (2)
        write (0, '(a48)') 'BC type (x = Lx):  Maxwell''s accommodation model'
        if ( (BC%p(2)<0.d0) .OR. (BC%p(2)>1.d0) ) stop 'ERROR: inappropriate accommodation coefficient specified for x = Lx'
        write (0, '(a28,es10.3)') 'Accomodation coefficient:   ', BC%p(2)
      case default
        write (0, '(a48)') '------------------------------------------------'
        stop "ERROR: incorrect boundary condition type specified"
      end select
      write (0, '(a28,es10.3)') 'y-velocity:                 ', BC%u(2,2)
      write (0, '(a28,es10.3)') 'z-velocity:                 ', BC%u(2,3)
      write (0, '(a28,es10.3)') 'temperature:                ', BC%c(2)
      write (0, '(a48)') '------------------------------------------------'

!     simulation progress header
      write (0, *)
      write (0, '(a41)') '[time,    particles]: simulation progress'
      write (0, '(a48)') '================================================'
      flush (0)

    end subroutine display_sim
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY TIME STAMP EVERY Nout STEPS
    subroutine display_rec (step, Nt, Np)
      implicit none
      integer :: step, Np, Nt
      real :: tcpu
      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a7,i9,a3,i9)') '[',tcpu,Np,']: step',step,' of',Nt
      flush (0)

    end subroutine display_rec
!----------------------------------------------------------------------------------------------------------------------------------

!   END SIMULATION
    subroutine end_sim (Np, bounds)
      integer :: Np
      type(Tbounds) :: bounds

!     update bounds factors
      open (2, file='bounds.inp')
      rewind (2)
      write (2,*) bounds%Ya_adv,  bounds%Yb_adv,  real(bounds%r_adv - 1.d0)
      write (2,*) bounds%Ya_coll, bounds%Yb_coll, real(bounds%r_coll - 1.d0)
      close (2)

!     display end time stamp
      call display_end (Np)

    end subroutine end_sim
!----------------------------------------------------------------------------------------------------------------------------------

!   EVALUATE MAXWELLIAN
    function eval_fMB (c, fMB)
      implicit none
      real(8) :: eval_fMB
      real(8), dimension(3) :: c
      type(TfMB) :: fMB

      eval_fMB = fMB%n / (srpi*fMB%c)**3 * exp ( - dot_product( c-fMB%u, c-fMB%u ) / fMB%c**2 )

    end function eval_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   INITIALIZE SIMULATION
    subroutine initialize_sim (cref, f0, bounds, part, sim, time)
      implicit none
      integer :: i
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tbounds) :: bounds
      type(Tpart) :: part
      type(Tsim) :: sim
      type(Ttime) :: time

!     Reset collision time counters
      time%tc = 0.d0
      time%dtc = -1.d0

!     reset number of simulated particles
      part%N = (/0,0/)

!     initialize collision mass conservation variable
      cref%DS = 0

!     initialization time stamp
      if (sim%disp>2) call display_init (sum(part%N))

    end subroutine initialize_sim
!----------------------------------------------------------------------------------------------------------------------------------

!   READ INPUT DATA AND FIX SIMULATION PARAMETERS
    subroutine input_data (BC, bref, cref, data, f0, floc, part, trans, steady, sim, time)
      use sampling_bound_constants
      implicit none
      integer :: i, cell
      real(8), dimension(:,:,:), allocatable :: data
      type(TBC) :: BC
      type(Tbref) :: bref
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tprop) :: trans, steady
      type(Tsim) :: sim
      type(Ttime) :: time

!     open input file
      open (1,file='input.inp',status='old')

!     read input data
      read (1,*) sim%Nt, sim%dt, sim%Nout, sim%Nss
      read (1,*) sim%Ncell, sim%Lx, sim%Xi, sim%Nens, sim%disp
      read (1,*) sim%Kn, sim%eps, sim%Pz, sim%Tz
      read (1,*) BC%type(1), BC%p(1), BC%u(1,2), BC%u(1,3), BC%c(1)
      read (1,*) BC%type(2), BC%p(2), BC%u(2,2), BC%u(2,3), BC%c(2)

!     read bound factor data
      open (2,file='bounds.inp',status='old')
      read (2,*) bounds%Ya_adv, bounds%Yb_adv, bounds%r_adv
      read (2,*) bounds%Ya_coll, bounds%Yb_coll, bounds%r_coll
      close (2)

!     set force logical variable
      if ( abs(sim%Pz) + abs(sim%Tz) > eps1 ) then
        sim%force = .true.
      else
        sim%force = .false.
      end if

!     reference equilibrium distribution
      f0%n = 1.d0
      f0%u = (/0.d0,0.d0,0.d0/)
      f0%c = 1.d0

!     allocate cell and particle arrays
      call allocate_arrays (bref, cref, data, floc, part, trans, steady, sim)

!     calculate cell spatial disctretization
      sim%dx = sim%Lx/sim%Ncell

!     boundary element corner
      BC%x(1) = 0.d0
      BC%x(2) = sim%Lx

!     inward surface normals
      BC%norm(1,:) = (/1,0,0/)
      BC%norm(2,:) = (/-1,0,0/)

!     remove normal velocity components
      BC%u(1,1) = 0.d0
      BC%u(2,1) = 0.d0

!     set corner coordinates
      do cell = 1, sim%Ncell
        cref%x(cell) = sim%dx*(cell-1)
      end do

!     create spatial coordinate data file
      open (23, file='x.dat', status='unknown')
      do cell = 1, sim%Ncell
        write (23,'(3es13.5)') cref%x(cell)+sim%dx/2
      end do
      close (23)

!     rescale time step to the advective time scale
      sim%dt = sim%dx*sim%dt/f0%c

!     display simulation parameters
      call display_sim (BC, sim)

!     calculate Neff, which is scaled by the perturbation parameter and reference density
      sim%Xi = sim%eps/sim%Xi

!     convert perturbed body forces to true body forces
      sim%Pz = sim%Pz * sim%eps
      sim%Tz = sim%Tz * sim%eps

!     convert perturbed boundary conditions to true boundary conditions
      do i = 1, 2
        BC%u(i,:) = sim%eps*BC%u(i,:)
        BC%c(i) = sqrt(1 + sim%eps*BC%c(i))
      end do

!     set boundary densities for reflections
      do i = 1, 2
        BC%n(i) = f0%n * (f0%c/BC%c(i))
      end do

!     set sampling bounds for body force generation term
      call set_force_bounds (f0, sim)

!     convert Knudsen number to dimensionless prefactor for collision integral
      sim%Kn = 2 / srpi / sim%Kn

!     force accomodation coefficients for specular boundary condtions to be zero
      do i = 1, 2
        if (BC%type(i)==0) BC%p(i) = 0.d0
      end do

!     set constants for sampling boundary generation term
      call set_advection_bounds (BC, f0, sim)

!     convert bound factors
      bounds%r_adv  = 1 + bounds%r_adv
      bounds%r_coll = 1 + bounds%r_coll

!     reset seed for random number generator
      call reset_seed

    end subroutine input_data
!----------------------------------------------------------------------------------------------------------------------------------

!   LOCATE A PARTICLE TO A PARTICULAR CELL OR BOUNDARY SURFACE
    subroutine locate (esc, cell, x, sim)
      use types
      implicit none
      logical :: esc
      integer :: cell
      real(8) :: x
      type(Tsim) :: sim

!     calculate cell number
      cell = 1 + floor( sim%Ncell * x / sim%Lx )

!     determine boundary cell, if particle escapes
      if ( cell < 1 ) then
        esc = .true.
        cell = 1
      else if ( cell > sim%Ncell ) then
        esc = .true.
        cell = 2
      else
        esc = .false.
      end if

    end subroutine locate
!----------------------------------------------------------------------------------------------------------------------------------

!   OUTPUT ENSEMBLE AVERAGED DATA
    subroutine output_prop_trans (data, sim)
      implicit none
      integer :: Ndt, i, j
      real(8) :: n, T, u2, eps
      real(8), dimension(3) :: u, q
      real(8), dimension(6) :: p
      real(8), dimension(:,:,:), allocatable :: data
      type(Tsim) :: sim

!     number of time steps
      Ndt = sim%Nt / sim%Nout

!     perturbation parameter
      eps = sim%eps

!     perform ensemble averaging on the moments of the velocity distribution
      data = data/sim%Nens

!     open data output file
      open (3, file='output.dat', status='unknown')

!     output ensemble averaged values
      do j = 1, Ndt
      do i = 1, sim%Ncell

!       construct dimensionless hydrodynamic properties from velocity moments
        n = sim%Xi * data(i,j,1)
        u = sim%Xi/(1.d0+n) * data(i,j,2:4)
        u2 = dot_product(u,u)
        p(1) = 2.d0*( sim%Xi*data(i,j,5) - (1.d0+n)*u(1)**2 )
        p(2) = 2.d0*( sim%Xi*data(i,j,6) - (1.d0+n)*u(1)*u(2) )
        p(3) = 2.d0*( sim%Xi*data(i,j,7) - (1.d0+n)*u(2)**2 )
        p(4) = 2.d0*( sim%Xi*data(i,j,8) - (1.d0+n)*u(2)*u(3) )
        p(5) = 2.d0*( sim%Xi*data(i,j,9) - (1.d0+n)*u(3)**2 )
        p(6) = 2.d0*( sim%Xi*data(i,j,10) - (1.d0+n)*u(3)*u(1) )
        T = (sim%Xi*data(i,j,11)/1.5d0 - n)/(1.d0+n) - u2/1.5d0
        q(1) = sim%Xi*data(i,j,12) - u(1) - p(1)*u(1) - p(2)*u(2) - p(6)*u(3) - (1.d0+n)*u(1)*(1.5d0*(1.d0+T)+u2)
        q(2) = sim%Xi*data(i,j,13) - u(2) - p(2)*u(1) - p(3)*u(2) - p(4)*u(3) - (1.d0+n)*u(2)*(1.5d0*(1.d0+T)+u2)
        q(3) = sim%Xi*data(i,j,14) - u(3) - p(6)*u(1) - p(4)*u(2) - p(5)*u(3) - (1.d0+n)*u(3)*(1.5d0*(1.d0+T)+u2)

!       scale properties based on eps
        n = n/eps
        u = u/eps
        p = p/eps
        T = T/eps
        q = q/eps

!       write values to file
        write (3,'(14es13.5)') n, u, p, T, q

      end do
      end do

!     close data output file
      close (3)

    end subroutine output_prop_trans
!----------------------------------------------------------------------------------------------------------------------------------

!   OUTPUT STEADY ENSEMBLE AVERAGED DATA
    subroutine output_prop_steady (steady, sim)
      implicit none
      integer :: Navg, i
      real(8) :: n, T, u2, eps
      real(8), dimension(3) :: u, q
      real(8), dimension(6) :: p
      type(Tprop) :: steady
      type(Tsim) :: sim

!     number of points averaged over
      Navg = sim%Nens*(sim%Nt+1-sim%Nss)

!     perturbation parameter
      eps = sim%eps

!     open data output file
      open (3, file='output.dat', status='unknown')

!     output ensemble averaged values
      do i = 1, sim%Ncell

!       perform ensemble averaging
        n = steady%n(i)/Navg
        u = steady%u(i,:)/Navg
        p = steady%p(i,:)/Navg
        T = steady%T(i)/Navg
        q = steady%q(i,:)/Navg

!       construct dimensionless hydrodynamic properties from velocity moments
        n = sim%Xi * n
        u = sim%Xi/(1.d0+n) * u
        u2 = dot_product(u,u)
        p(1) = 2.d0*( sim%Xi*p(1) - (1.d0+n)*u(1)**2 )
        p(2) = 2.d0*( sim%Xi*p(2) - (1.d0+n)*u(1)*u(2) )
        p(3) = 2.d0*( sim%Xi*p(3) - (1.d0+n)*u(2)**2 )
        p(4) = 2.d0*( sim%Xi*p(4) - (1.d0+n)*u(2)*u(3) )
        p(5) = 2.d0*( sim%Xi*p(5) - (1.d0+n)*u(3)**2 )
        p(6) = 2.d0*( sim%Xi*p(6) - (1.d0+n)*u(3)*u(1) )
        T = (sim%Xi*T/1.5d0 - n)/(1.d0+n) - u2/1.5d0
        q(1) = sim%Xi*q(1) - u(1) - p(1)*u(1) - p(2)*u(2) - p(6)*u(3) - (1.d0+n)*u(1)*(1.5d0*(1.d0+T)+u2)
        q(2) = sim%Xi*q(2) - u(2) - p(2)*u(1) - p(3)*u(2) - p(4)*u(3) - (1.d0+n)*u(2)*(1.5d0*(1.d0+T)+u2)
        q(3) = sim%Xi*q(3) - u(3) - p(6)*u(1) - p(4)*u(2) - p(5)*u(3) - (1.d0+n)*u(3)*(1.5d0*(1.d0+T)+u2)

!       scale properties based on eps
        n = n/eps
        u = u/eps
        p = p/eps
        T = T/eps
        q = q/eps

!       write values to file
        write (3,'(14es13.5)') n, u, p, T, q
      end do

!     close data output file
      close (3)

    end subroutine output_prop_steady
!----------------------------------------------------------------------------------------------------------------------------------

!   UPDATE PROPERTIES
    subroutine properties (f0, part, trans, sim)
      implicit none
      integer :: i, p, s
      real(8) :: c2
      real(8), dimension(3) :: c
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     zero out properties
      do i = 1, sim%Ncell
        trans%n(i) = 0.d0
        trans%u(i,:) = (/0.d0,0.d0,0.d0/)
        trans%T(i) = 0.d0
        trans%p(i,:) = (/0.d0,0.d0,0.d0,0.d0,0.d0,0.d0/)
        trans%q(i,:) = (/0.d0,0.d0,0.d0/)
      end do

!     sum the moments corresponding to density, velocity, pressure tensor, temperature, and heat flux

!     positive particles
      do p=1,part%N(0)
        i = part%cell(0,p)
        c = (part%c(0,p,:) - f0%u)/f0%c
        trans%n(i) = trans%n(i) - 1
        trans%u(i,:) = trans%u(i,:) - c
        trans%p(i,1) = trans%p(i,1) - c(1)**2
        trans%p(i,2) = trans%p(i,2) - c(1)*c(2)
        trans%p(i,3) = trans%p(i,3) - c(2)**2
        trans%p(i,4) = trans%p(i,4) - c(2)*c(3)
        trans%p(i,5) = trans%p(i,5) - c(3)**2
        trans%p(i,6) = trans%p(i,6) - c(3)*c(1)
        c2 = dot_product(c,c)
        trans%T(i) = trans%T(i) - c2
        trans%q(i,:) = trans%q(i,:) - part%c(0,p,:)*c2
      end do

!     negative particles
      do p=1,part%N(1)
        i = part%cell(1,p)
        c = (part%c(1,p,:) - f0%u)/f0%c
        trans%n(i) = trans%n(i) + 1
        trans%u(i,:) = trans%u(i,:) + c
        trans%p(i,1) = trans%p(i,1) + c(1)**2
        trans%p(i,2) = trans%p(i,2) + c(1)*c(2)
        trans%p(i,3) = trans%p(i,3) + c(2)**2
        trans%p(i,4) = trans%p(i,4) + c(2)*c(3)
        trans%p(i,5) = trans%p(i,5) + c(3)**2
        trans%p(i,6) = trans%p(i,6) + c(3)*c(1)
        c2 = dot_product(c,c)
        trans%T(i) = trans%T(i) + c2
        trans%q(i,:) = trans%q(i,:) + part%c(1,p,:)*c2
      end do

    end subroutine properties
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A RANDOM INTEGER FROM 1 TO N
    function rand_int(N)
      implicit none
      integer :: N, rand_int
      real(8) :: r

      call random_number(r)
      rand_int = int(r*N) + 1

    end function rand_int
!----------------------------------------------------------------------------------------------------------------------------------

!   SAVE DATA TO BINARY FILE FOR ENSEMBLE AVERAGING
    subroutine record_prop_trans (step, Ncell, data, trans)
      implicit none
      integer :: step, Ncell, i
      real(8), dimension(:,:,:) :: data
      type(Tprop) :: trans

      do i = 1, sim%Ncell
        data(i,step,1)     = data(i,step,1)     + trans%n(i)
        data(i,step,2:4)   = data(i,step,2:4)   + trans%u(i,:)
        data(i,step,5:10)  = data(i,step,5:10)  + trans%p(i,:)
        data(i,step,11)    = data(i,step,11)    + trans%T(i)
        data(i,step,12:14) = data(i,step,12:14) + trans%q(i,:)
      end do

    end subroutine record_prop_trans
!----------------------------------------------------------------------------------------------------------------------------------

!   SAVE STEADY STATE PROPERTY DATA FOR ENSEMBLE AVERAGING
    subroutine record_prop_steady (trans, steady, sim)
      implicit none
      integer :: i
      type (Tprop) :: trans, steady
      type (Tsim) :: sim

!     cell-based properties
      do i = 1, sim%Ncell
        steady%n(i) = steady%n(i) + trans%n(i)
        steady%u(i,:) = steady%u(i,:) + trans%u(i,:)
        steady%p(i,:) = steady%p(i,:) + trans%p(i,:)
        steady%T(i) = steady%T(i) + trans%T(i)
        steady%q(i,:) = steady%q(i,:) + trans%q(i,:)
      end do

    end subroutine record_prop_steady
!----------------------------------------------------------------------------------------------------------------------------------

!   RESET SEED FOR RANDOM NUMBER GENERATOR TO SYSTEM DATE AND TIME DERIVED VALUE
    subroutine reset_seed
      implicit none
      integer :: i_seed
      integer, dimension(:), allocatable :: a_seed
      integer, dimension(1:8) :: dt_seed

      call random_seed (size = i_seed)
      allocate (a_seed(1:i_seed))
      call random_seed (get = a_seed)
      call date_and_time (values = dt_seed)
      a_seed(i_seed) = dt_seed(8); a_seed(1) = dt_seed(8) * dt_seed(7) * dt_seed(6)
      call random_seed (put = a_seed)
      deallocate (a_seed)

    end subroutine reset_seed
!----------------------------------------------------------------------------------------------------------------------------------

!   SET BOUNDARY GENERATION BOUNDS BASED ON LINEARIZED CONDITIONS
    subroutine set_advection_bounds (BC, f0, sim)
      use numerical_constants
      use types
      use sampling_bound_constants
      implicit none
      integer :: i
      real(8) :: na, dn, ca, dc, acc
      real(8), dimension(3) :: du
      type(TBC) :: BC
      type(TfMB) :: f0
      type(Tsim) :: sim

      do i=1,2

!       for specular boundary elements, prevent particle generation and cycle to next boundary element
        BC%Ns(i) = 1.d0
        if (BC%type(i)==0) then
          BC%Ns(i) = 0.d0
          cycle
        end if

!       compute average properties for generating particles at the boundaries
        na = (f0%n + BC%n(i))/2
        ca = (f0%c + BC%c(i))/2

!       compute the absolute property differences for estimating the sampling bounds
        dn = (BC%n(i) - f0%n)/na
        du = BC%u(i,:) / ca
        dc = (BC%c(i) - f0%c)/ca

!       compute linearized sampling bounds
        BC%a(i) = ( abs(dn-3.d0*dc)*an + 2.d0*abs(du(1))*ann + 2.d0*abs(du(2))*ans + 2.d0*abs(du(3))*ans + 2.d0*abs(dc)*a2n )**0.4
        BC%b(i,1)   = ( abs(dn-3.d0*dc)*bn_n + 2.d0*abs(du(1))*bnn_n + 2.d0*abs(du(2))*bns_n + 2*abs(du(3))*bns_n                 &
                                                                                                        + 2.d0*abs(dc)*b2n_n )**0.2
        BC%b(i,2)   = ( abs(dn-3.d0*dc)*bn_s + 2.d0*abs(du(1))*bns_n + 2.d0*abs(du(2))*bns_n + 2*abs(du(3))*bns_t                 &
                                                                                                        + 2.d0*abs(dc)*b2n_s )**0.2
        BC%b(i,3)   = ( abs(dn-3.d0*dc)*bn_s + 2.d0*abs(du(1))*bns_n + 2.d0*abs(du(2))*bns_t + 2*abs(du(3))*bns_n                 &
                                                                                                        + 2.d0*abs(dc)*b2n_s )**0.2

!       compute number of trial samples per time step dt
        BC%Ns(i) = 10.d0/srpi**3 * na*ca * 1.d0/sim%Xi/sim%dx * BC%a(i)*product(BC%b(i,:)) * BC%Ns(i)

!       adjust for accomodating boundary condtion type
        if (BC%type(i)==2) BC%Ns(i) = BC%Ns(i) * BC%p(i)

!       store average properties
        BC%na(i) = na
        BC%ca(i) = ca
      end do

    end subroutine set_advection_bounds
!----------------------------------------------------------------------------------------------------------------------------------

!   SET BOUNDARY GENERATION BOUNDS BASED ON LINEARIZED CONDITIONS
    subroutine set_force_bounds (f0, sim)
      use numerical_constants
      use types
      use sampling_bound_constants
      implicit none
      type(TfMB) :: f0
      type(Tsim) :: sim

!     set sampling bounds for body force generation term
      sim%af = ( an*abs(sim%Pz) + (2.5d0*an+a2n)*abs(sim%Tz) )**0.4
      sim%bf(1) = ( bn_s*abs(sim%Pz) + (2.5d0*bn_s+b2n_s)*abs(sim%Tz) )**0.2
      sim%bf(2) = ( bn_s*abs(sim%Pz) + (2.5d0*bn_s+b2n_s)*abs(sim%Tz) )**0.2
      sim%bf(3) = ( bn_n*abs(sim%Pz) + (2.5d0*bn_n+b2n_n)*abs(sim%Tz) )**0.2
      sim%Nf = 20.d0 * f0%n/(srpi*f0%c)**3 *sim%af*product(sim%bf) * (sim%dt/2) * sim%Lx/sim%dx/sim%Xi

    end subroutine set_force_bounds
!----------------------------------------------------------------------------------------------------------------------------------

!   SHUFFLE THE ELEMENTS IN A VECTOR OF LENGTH N
    subroutine shuffle (X, N)
      implicit none
      integer :: N, i, j, Xs
      integer, dimension(:) :: X

      do i=1,N
        j = rand_int(N+1-i) + (i-1)
        Xs = X(j)
        X(j) = X(i)
        X(i) = Xs
      end do

    end subroutine shuffle
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0,*)
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

!   ADVANCE SIMULATION BY ONE TIME STEP
!   second-order time splitting similar to Strang's method for DSMC [see Rader, et. al. Phys. Fluids (2006)]
    subroutine time_step (BC, bref, f0, floc, bounds, part, trans, steady, sim, step, time)
      implicit none
      integer :: step
      type (TBC) :: BC
      type (Tbref) :: bref
      type (TfMB) :: f0
      type (Tfloc) :: floc
      type (Tbounds) :: bounds
      type (Tpart) :: part
      type (Tprop) :: trans, steady
      type (Tsim) :: sim
      type (Ttime) :: time

!     perform intial advection half step
      call advection (BC, bref, f0, bounds, part, trans, sim)

!     perform effective "body force" half step
      if (sim%force) call force (f0, part, sim)

!     perform full collision step
      call collision (bounds, cref, f0, floc, part, trans, sim, time)

!     perform effective "body force" half step
      if (sim%force) call force (f0, part, sim)

!     perform final half advection step
      call advection (BC, bref, f0, bounds, part, trans, sim)

    end subroutine time_step
!----------------------------------------------------------------------------------------------------------------------------------

  end program LVDSMC


!==================================================================================================================================
! ADVECTION ROUTINE
!==================================================================================================================================

! ADVECTION ROUTINE FOR HALF TIME STEP
  subroutine advection (BC, bref, f0, bounds, part, trans, sim)
    use types
    use numerical_constants
    implicit none
    logical :: esc
    integer :: i, j, cell, type
    integer, dimension(0:1) :: NG
    real(8) :: dt, r, dtf
    type(TBC) :: BC
    type(Tbref) :: bref
    type(TfMB) :: f0
    type(Tbounds) :: bounds
    type(Tpart) :: part
    type(Tprop) :: trans
    type(Tsim) :: sim

!   time step
    dt = sim%dt/2

!   clear boundary cell index data
    do i = 1, 2
      bref%N(:,i) = (/0,0/);
    end do

!   advect particles along velocity trajectories
    do i = 0, 1
    do j = 1, part%N(i)
      part%x(i,j) = part%x(i,j) + dt * part%c(i,j,1)
      call locate (esc, cell, part%x(i,j), sim)
      part%cell(i,j) = cell

!     enforce boundary condition for escaping particles
      if (esc) then

!       determine boundary condition type
        type = BC%type(cell)
        if (type==2) then
          call random_number(r)
          if (r<BC%p(cell)) then
            type = 1
          else
            type = 0
          end if
        end if
        select case (type)

!       specular reflection
        case (0)
          dtf = dt
          call specular_reflection (i, j, dtf, BC, part, trans, sim)

!       diffuse reflection
        case (1)

!         warn if the maximum number of particles per boundary is exceeded
          if (bref%N(i,cell) >= Nbmax) call stop_exceed_Nbmax (sum(part%N(:)))

!         create lists of positive and negative particles for cancelling and reflection
          bref%N(i,cell) = bref%N(i,cell) + 1
          bref%ref(i,cell,bref%N(i,cell)) = j
        end select

!     update particle cell location if it remains within the simulation domain
      else
        part%cell(i,j) = cell
      end if
    end do
    end do

!   mark pairs of positive and negative particles from diffuse reflections for deletion
    dtf = dt
    call boundary_cancel (dtf, BC, bref, part, sim)

!   removed deleted particles from simulation
    do i = 0, 1
    do j = part%N(i), 1, -1
      if (part%del(i,j)) call remove(i, j, part)
    end do
    end do

!   generate particles at the boundary
    do i = 1, 2

!     skip routine for specular reflections
      if (BC%type(i)==0) then

!     advection particle generation routine with mass conservation
      else
        call generate_boundary (i, dt, 0.5d0, BC, f0, bounds, part, trans, sim, NG, .false.)
        j = NG(0)
        NG(0) = NG(1)
        NG(1) = j
        call generate_boundary (i, dt, 0.5d0, BC, f0, bounds, part, trans, sim, NG, .true.)
      end if
    end do

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD PARTICLE TO SIMULATION
    subroutine add (c, x, s, cell, part)
      implicit none
      integer :: s, cell
      real(8) :: x
      real(8), dimension(3) :: c
      type(Tpart) :: part

!     warn if the maximum number of particles is exceeded
      if (maxval(part%N)>=Nmax) call stop_exceed_Nmax (sum(part%N))

!     add new velocity sample
      part%N(s) = part%N(s) + 1
      part%c(s, part%N(s),:) = c
      part%x(s, part%N(s)) = x
      part%cell(s, part%N(s)) = cell
      part%del(s, part%N(s)) = .false.

    end subroutine add
!----------------------------------------------------------------------------------------------------------------------------------

!   CANCEL PAIRS OF POSITIVE AND NEGATIVE PARTICLES ESCAPING THROUGH THE BOUNDARY
    subroutine boundary_cancel (dt, BC, bref, part, sim)
      implicit none
      integer :: i, j, k, s
      integer, dimension(:), allocatable :: list
      real(8) :: dt, dtf
      type(TBC) :: BC
      type(Tbref) :: bref
      type(Tpart) :: part
      type(Tsim) :: sim

!     perform cancellation for each boundary of type 1 or 2
      do i = 1, 2
        if ((BC%type(i)==1).OR.(BC%type(i)==2)) then

!         delete all reflecting particles if Np = Nn
          if (bref%N(0,i)==bref%N(1,i)) then
            do j = 1, bref%N(0,i)
              part%del(0,bref%ref(0,i,j)) = .true.
              part%del(1,bref%ref(1,i,j)) = .true.
            end do

!         case: Np > Nn
          else if (bref%N(0,i) < bref%N(1,i)) then

!           create list of positive particles to delete
            allocate (list(bref%N(1,i)))
            list = bref%ref(1, i, 1:bref%N(1,i))
            call shuffle (list, bref%N(1,i))

!           mark particles for deletion
            do j = 1, bref%N(0,i)
              part%del(0, bref%ref(0,i,j)) = .true.
              part%del(1, list(j)) = .true.
            end do

!           diffusively reflect remaining particles
            do j = bref%N(0,i)+1, bref%N(1,i)
              dtf = dt
              call diffuse_reflection (1, list(j), dtf, BC, part, trans, sim)
            end do

!           deallocate list array
            deallocate (list)

!         case: Nn > Np
          else

!           create list of negative particles to delete
            allocate (list(bref%N(0,i)))
            list = bref%ref(0, i, 1:bref%N(0,i))
            call shuffle (list, bref%N(0,i))

!           mark particles for deletion
            do j = 1, bref%N(1,i)
              part%del(0, list(j)) = .true.
              part%del(1, bref%ref(1,i,j)) = .true.
            end do

!           diffusively reflect remaining particles
            do j=bref%N(1,i)+1, bref%N(0,i)
              dtf = dt
              call diffuse_reflection (0, list(j), dtf, BC, part, trans, sim)
            end do

!           deallocate list array
            deallocate (list)
          end if
        end if
      end do

    end subroutine boundary_cancel
!----------------------------------------------------------------------------------------------------------------------------------

!   DRAWS TWO NORMALLY DISTRIBUTED RANDOM VARIABLES USING THE POLAR FORM OF THE BOX MULLER TRANSFORMATION
    subroutine box_Muller (x,y)
      implicit none
      real(8) :: x, y, z

      do
        call random_number(x)
        call random_number(y)
        x=2*x-1
        y=2*y-1
        z=x*x+y*y
        if (z<1.d0) exit
      end do
      z=sqrt((-2*log(z))/z)
      x=sqrt(0.5d0)*x*z
      y=sqrt(0.5d0)*y*z

    end subroutine box_Muller
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM DIFFUSE REFLECTION
    subroutine diffuse_reflection (s, i, dt, BC, part, trans, sim)
      implicit none
      logical :: esc
      integer :: s, i, cell, type, dir
      real(8) :: dt, r, xb
      real(8), dimension(3) :: c
      type(TBC) :: BC
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     boundary element to reflect from
      cell = part%cell(s,i)

!     reflected particle direction
      if ( cell == 1 ) then
        dir = 1
      else
        dir = -1
      end if

!     determine intersection point and remaining time step
      call intersection (cell, xb, part%x(s,i), part%c(s,i,:), dt, sim)

!     sample new velocity and advect to new position
      call sample_fMB_n (c, dir, BC%u(part%cell(s,i),:), BC%c(part%cell(s,i)))
      part%c(s,i,:) = c
      part%x(s,i) = xb + c(1)*dt

!     determine new cell location
      call locate (esc, cell, part%x(s,i), sim)
      part%cell(s,i) = cell

    end subroutine diffuse_reflection
!----------------------------------------------------------------------------------------------------------------------------------

!   EVALUATE MAXWELLIAN
    function eval_fMB (c, fMB)
      implicit none
      real(8) :: eval_fMB
      real(8), dimension(3) :: c
      type(TfMB) :: fMB

      eval_fMB = fMB%n / (srpi*fMB%c)**3 * exp ( - dot_product( c-fMB%u, c-fMB%u ) / fMB%c**2 )

    end function eval_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   GENERATE ADDITIONAL PARTICLES DUE TO ADVECTION, NONLINEAR VERSION
    subroutine generate_boundary (i, dt, frac, BC, f0, bounds, part, trans, sim, NG, opt)
      implicit none
      logical :: esc, opt, sgn
      integer :: i, N, p, cell, type, s
      integer, dimension(0:1) :: NG
      real(8) :: r, f, H, dt, dtf, a, trial, frac, x
      real(8), dimension(3) :: V, z, c, b
      type(TBC) :: BC
      type(TfMB) :: f0, fb
      type(Tbounds) :: bounds
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     boundary distribution
      fb%n = BC%n(i)
      fb%u = BC%u(i,:)
      fb%c = BC%c(i)

!     apply bound factors
      a    = BC%a(i)   * bounds%Ya_adv    * bounds%r_adv
      b(1) = BC%b(i,1) * bounds%Yb_adv(1) * bounds%r_adv
      b(2) = BC%b(i,2) * bounds%Yb_adv(2) * bounds%r_adv
      b(3) = BC%b(i,3) * bounds%Yb_adv(2) * bounds%r_adv

!     use number of trial particles as exit condition when opt=.false.
      if (.NOT.opt) then

!       determine the number of trial particles
        trial = dt*frac*BC%Ns(i) * bounds%Ya_adv * bounds%Yb_adv(1) * bounds%Yb_adv(2)**2 * bounds%r_adv**4
        N = int(trial)
        call random_number(r)
        if (r < trial-N) N = N + 1

!       initialize the number of generated particles
        NG = (/0,0/)
      else

!       invert the sign of the desired number of particles to generate
        NG = -NG
      end if

!     generate particles
      p = 0
      do

!       exit criterion
        if (opt) then
          if (sum(NG)==0) exit
        else
          if (p==N) exit
        end if

!       advance do loop counter
        p = p + 1

!       generate random deviates
        call random_number(r)
        H = r*a
        call random_number(r)
        V(1) = r*b(1)
        call random_number(r)
        V(2) = (2.d0*r-1.d0)*b(2)
        call random_number(r)
        V(3) = (2.d0*r-1.d0)*b(3)

!       velocity
        z = V/sqrt(H)
        c = BC%ca(i)*z + 0.5d0*BC%u(i,:)

!       evaluate distribution
        f = BC%ca(i)**2/BC%na(i) * srpi**3 * c(1) * ( eval_fMB(c,fb) - eval_fMB(c,f0) )

!       particle sign
        s = sign_index(f)

!       acceptance criteria
        sgn = NG(s) /= 0
        if (abs(f)**0.4d0 > H) then
        if ( (.NOT.opt) .OR. sgn ) then

!         transform velocity to global coordinates
          if ( i == 2 ) c(1) = - c(1)

!         advect a random fraction of the time step away from the interface
          call random_number(r)
          dtf = r*dt
          x = BC%x(i) + c(1)*dtf

!         determine cell number
          call locate (esc, cell, x, sim)

!         add particle to simulation
          call add (c, x, s, cell, part)
          NG(s) = NG(s) + 1

!         update bound factors if necessary
          call update_bounds (H, V, a, b, bounds, sum(part%N))
        end if
        end if
      end do

    end subroutine generate_boundary
!----------------------------------------------------------------------------------------------------------------------------------

!   DETERMINE INTERSECTION POINT OF BOUNDARY AND PARTICLE TRAJECTORY
    subroutine intersection (b, xb, x, c, dt, sim)
      use types
      implicit none
      integer :: b
      real(8) :: dt, x, xb
      real(8), dimension(3) :: c
      type(Tsim) :: sim

      select case (b)

!     for x=0 boundary
      case (1)
        xb = 0.d0
        dt = x / c(1)

!     for x=Lx boundary
      case (2)
        xb = sim%Lx
        dt = ( x - sim%Lx ) / c(1)

      end select

    end subroutine intersection
!----------------------------------------------------------------------------------------------------------------------------------

!   LOCATE A PARTICLE TO A PARTICULAR CELL OR BOUNDARY SURFACE
    subroutine locate (esc, cell, x, sim)
      use types
      implicit none
      logical :: esc
      integer :: cell
      real(8) :: x
      type(Tsim) :: sim

!     calculate cell number
      cell = 1 + floor( sim%Ncell * x / sim%Lx )

!     determine boundary cell, if particle escapes
      if ( cell < 1 ) then
        esc = .true.
        cell = 1
      else if ( cell > sim%Ncell ) then
        esc = .true.
        cell = 2
      else
        esc = .false.
      end if

    end subroutine locate
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A RANDOM INTEGER FROM 1 TO N
    function rand_int(N)
      implicit none
      integer :: N, rand_int
      real(8) :: r

      call random_number(r)
      rand_int = int(r*N) + 1

    end function rand_int
!----------------------------------------------------------------------------------------------------------------------------------

!   REMOVE A PARTICLE FROM SIMULATION
    subroutine remove (i, j, part)
      implicit none
      integer :: i, j
      type(Tpart) :: part

!     remove particle
      part%x(i,j) = part%x(i,part%N(i))
      part%c(i,j,:) = part%c(i,part%N(i),:)
      part%cell(i,j) = part%cell(i,part%N(i))
      part%del(i,j) = .false.
      part%N(i) = part%N(i) - 1

    end subroutine remove
!----------------------------------------------------------------------------------------------------------------------------------

!   SAMPLE FROM FLUXAL MAXWELL-BOLTZMANN DISTRIBUTION
    subroutine sample_fMB_n (c, dir, u, v)
      implicit none
      integer :: dir
      real(8) :: v, r
      real(8), dimension(3) :: c, u

!     sample from fMB in tangential directions
      call box_Muller (c(2),c(3))

!     sample from fluxal distribution in normal direction
      call random_number(r)
      c(1) = dir * sqrt( -log(r) )

!     transform to global coordinates
      c = u + v*c

    end subroutine sample_fMB_n
!----------------------------------------------------------------------------------------------------------------------------------

!   SHUFFLE THE ELEMENTS IN A VECTOR OF LENGTH N
    subroutine shuffle (X, N)
      implicit none
      integer :: N, i, j, Xs
      integer, dimension(:) :: X

      do i=1,N
        j = rand_int(N+1-i) + (i-1)
        Xs = X(j)
        X(j) = X(i)
        X(i) = Xs
      end do

    end subroutine shuffle
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM DIFFUSE REFLECTION
    subroutine specular_reflection (i, j, dt, BC, part, trans, sim)
      implicit none
      logical :: esc
      integer :: i, j, cell, type
      integer, dimension(3) :: n
      real(8) :: dt, r, xb
      type(TBC) :: BC
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     boundary element to reflect from
      cell = part%cell(i,j)

!     inward boundary normal
      n = BC%norm(cell,:)

!     determine intersection point and remaining time step
      call intersection (cell, xb, part%x(i,j), part%c(i,j,:), dt, sim)

!     find new velocity and advect to new position
      part%c(i,j,:) = part%c(i,j,:) - 2*dot_product(part%c(i,j,:),n)*n
      part%x(i,j) = xb + part%c(i,j,1)*dt

!     determine new cell location
      call locate (esc, cell, part%x(i,j), sim)
      part%cell(i,j) = cell

    end subroutine specular_reflection
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0,*)
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES PER BOUNDARY WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nbmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a50)') '[',tcpu,Np,']: ERROR: Nbmax exceeded--increase Nbmax parameter'
      write (0,*)
      flush (0)
      stop

    end subroutine stop_exceed_Nbmax
!----------------------------------------------------------------------------------------------------------------------------------

!   UPDATE BOUND FACTORS IF NECESSARY
    subroutine update_bounds (H, V, a, b, bounds, Np)
      implicit none
      logical :: change
      integer :: Np, disp
      real(8) :: H, a, tcpu, Vmax
      real(8), dimension(3) :: V, b
      type(Tbounds) :: bounds

      change = .false.

!     test bound on dependent variable
      if (H > a/bounds%r_adv) then
        bounds%Ya_adv = (H/a) * bounds%Ya_adv * bounds%r_adv
        change = .true.
      end if

!     test bound on independent variable, normal component
      if (V(1) > b(1)/bounds%r_adv) then
        bounds%Yb_adv(1) = (V(1)/b(1)) * bounds%Yb_adv(1) * bounds%r_adv
        change = .true.
      end if

!     test bound on independent variable, tangential component
      Vmax = max( abs(V(2)/b(2)), abs(V(3)/b(3)) )
      if (Vmax > 1.d0/bounds%r_adv) then
        bounds%Yb_adv(2) = Vmax * bounds%Yb_adv(2) * bounds%r_adv
        change = .true.
      end if

!     warn user if bound factors were updated
      if (change) then
        call cpu_time(tcpu)
        write (0, '(a55)') '-------------------------------------------------------'
        write (0, '(a1,es9.3,i9,a36)') '[',tcpu,Np,']: ATTENTION: bound factors changed!'
        write (0, '(a55)') '-------------------------------------------------------'
        flush (0)
      end if

    end subroutine update_bounds
!----------------------------------------------------------------------------------------------------------------------------------

!   GENERATE UNIFORMLY DISTRIBUTED SPATIAL POSITION WITHIN A CELL
    function uniform_x (cell, x0, dx)
      implicit none
      integer :: cell
      real(8) :: r, x0, dx, uniform_x

      call random_number(r)
      uniform_x = x0 + r*dx

    end function uniform_x
!----------------------------------------------------------------------------------------------------------------------------------

  end subroutine advection


!==================================================================================================================================
! COLLISION ROUTINE
!==================================================================================================================================

! BGK COLLISION ROUTINE
  subroutine collision (bounds, cref, f0, floc, part, prop, sim, time)
    use numerical_constants
    use types
    implicit none
    logical coll, first, flag
    integer :: cell, s, i
    real(8) :: r
    type(Tbounds) :: bounds
    type(Tcref) :: cref
    type(TfMB) :: f0
    type(Tfloc) :: floc
    type(Tpart) :: part
    type(Tprop) :: prop
    type(Tsim) :: sim
    type(Ttime) :: time

!   create cell based cross referencing
    call cross_ref (sim%Ncell, cref, floc, part)

!   update collision sampling bounds
    call set_collision_bounds (bounds, cref, f0, floc, sim)

!   perform collision jump events until the collision time step reaches the advection step
    first = .true.
    coll = .false.
    flag = .false.
    do

!     collision time step
      call step (first, coll, cref, floc, part, time, sim)
      if (coll) exit
      first = .false.
      flag = .false.

!     choose cell based on index distribution
      call choose_cell (cell, cref%J, sim%Ncell, cref%Jmax)

!     decide between trial generation and deletion steps
      call random_number(r)
      if ( r < cref%F(cell) / ( cref%F(cell) + sum(cref%N(:,cell)) ) ) then

!       trial particle creation step
        call create (cell, bounds, cref, f0, floc, part, sim, 0, flag)
      else

!       particle deletion step
        call choose_part (s, i, cell, cref)
        call remove_ref (s, i, cref, floc, part, sim)
        flag = .true.
      end if

!     update ratio-of-uniforms bounds
      if (flag) then
        call change_collision_bound (cell, bounds, cref, f0, floc, sim)
        flag = .false.
      end if
    end do

!   correct total sign parity
    call parity (bounds, cref, f0, floc, part, sim)

!   correct mass by resampling particles
    call regenerate (bounds, cref, f0, floc, part, sim)

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD PARTICLE TO SIMULATION MAINTAINING CROSS REFERENCING
    subroutine add_ref (c, x, s, cell, cref, f0, floc, part, sim)
      implicit none
      integer :: s, cell, ss
      real(8) :: x, Tloc
      real(8), dimension(3) :: c
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tsim) :: sim

!     numerical sign
      ss = 2*s-1

!     warn if the maximum number of particles is exceeded
      if (part%N(s) >= Nmax) call stop_exceed_Nmax (sum(part%N(:)))

!     add new velocity sample
      part%N(s) = part%N(s) + 1
      part%c(s,part%N(s),:) = c
      part%x(s,part%N(s)) = x
      part%cell(s,part%N(s)) = cell
      part%del(s,part%N(s)) = .false.
      cref%G(s,part%N(s)) = .true.

!     warn if the maximum number of particles per cell is exceeded
      if (cref%N(s,cell) >= Ncmax) call stop_exceed_Ncmax (sum(part%N(:)))

!     update cross referencing
      cref%N(s,cell) = cref%N(s,cell) + 1
      cref%ref(s,cell,cref%N(s,cell)) = part%N(s)
      cref%idx(s,part%N(s)) = cref%N(s,cell)

!     update local properties in affected cell
      cref%S(cell)   = cref%S(cell) + ss
      cref%U(cell,:) = cref%U(cell,:) + ss*c
      cref%C(cell)   = cref%C(cell) + ss*dot_product(c,c)
      floc%n(cell)   = 1 + sim%Xi * cref%S(cell)
      floc%u(cell,:) = sim%Xi * cref%U(cell,:) / floc%n(cell)
      Tloc           = ( 1 + sim%Xi * cref%C(cell) / 1.5d0 ) / floc%n(cell) - dot_product( floc%u(cell,:), floc%u(cell,:) ) / 1.5d0

!     determine local most probable molecular speed if temperature is positive
      if ( Tloc <= 0.d0 ) then

!       stop simulation if a negative temperature is encountered
        call stop_Tneg (sum(part%N))
      else

!       update most probable molecular speed
        floc%c(cell)   = sqrt( Tloc )
      end if

!     update cell index distribution
      cref%Jsum = cref%Jsum - cref%J(cell)
      cref%J(cell) = floc%n(cell) * ( cref%F(cell) + sum(cref%N(:,cell)) )
      cref%Jsum = cref%Jsum + cref%J(cell)
      if ( cref%J(cell) > cref%Jmax ) cref%Jmax = cref%J(cell)

!     update conservation enforcement tracking
      ss = 2*s-1
      cref%DS = cref%DS + ss
      cref%NG(s) = cref%NG(s) + 1
      cref%Glist(s,1,cref%NG(s)) = part%N(s)
      cref%Gidx(s,part%N(s)) = cref%NG(s)

    end subroutine add_ref
!----------------------------------------------------------------------------------------------------------------------------------

!   SET COLLISION GENERATION BOUNDS BASED ON LINEARIZED CONDITIONS
    subroutine change_collision_bound (i, bounds, cref, f0, floc, sim)
      use numerical_constants
      use types
      use sampling_bound_constants
      implicit none
      integer :: i
      real(8) :: na, dn, ca, dc
      real(8), dimension(3) :: ua, du
      type(Tbounds) :: bounds
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tfloc) :: floc
      type(Tsim) :: sim

!     remove old index distribution component
      cref%Jsum = cref%Jsum - cref%J(i)

!     compute average properties for generating particles at the boundaries
      na = (f0%n + floc%n(i))/2
      ua = (f0%u + floc%u(i,:))/2
      ca = (f0%c + floc%c(i))/2

!     compute the absolute property differences for estimating the sampling bounds
      dn = (floc%n(i) - f0%n) / na
      du = (floc%u(i,:) -f0%u) / ca
      dc = (floc%c(i) - f0%c) / ca

!     compute linearized sampling bounds
      cref%a(i) = ( abs(dn-3.d0*dc)*a0 + 2.d0*abs(du(1))*an + 2.d0*abs(du(2))*an + 2.d0*abs(du(3))*an + 2.d0*abs(dc)*a2 )**0.4
      cref%b(i,1) = ( abs(dn-3.d0*dc)*b0 + 2.d0*abs(du(1))*bn_n + 2.d0*abs(du(2))*bn_s + 2.d0*abs(du(3))*bn_s                     &
                                                                                                           + 2.d0*abs(dc)*b2 )**0.2
      cref%b(i,2) = ( abs(dn-3.d0*dc)*b0 + 2.d0*abs(du(1))*bn_s + 2.d0*abs(du(2))*bn_n + 2.d0*abs(du(3))*bn_s                     &
                                                                                                           + 2.d0*abs(dc)*b2 )**0.2
      cref%b(i,3) = ( abs(dn-3.d0*dc)*b0 + 2.d0*abs(du(1))*bn_s + 2.d0*abs(du(2))*bn_s + 2.d0*abs(du(3))*bn_n                     &
                                                                                                           + 2.d0*abs(dc)*b2 )**0.2

!     integral bound on difference between local and global equilibrium state
      cref%F(i) = 20.d0/srpi**3 * na/sim%Xi * cref%a(i)*product(cref%b(i,:)) * bounds%Ya_coll*bounds%Yb_coll**3*bounds%r_coll**4

!     store average properties
      cref%na(i)   = na
      cref%ua(i,:) = ua
      cref%ca(i)   = ca

!     change cell index distribution component
      cref%J(i) = floc%n(i) * ( cref%F(i) + sum(cref%N(:,i)) )
      cref%Jsum = cref%Jsum + cref%J(i)
      if ( cref%J(i) > cref%Jmax ) cref%Jmax = cref%J(i)

    end subroutine change_collision_bound
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE A CELL INDEX PROPORTIONAL TO DISCRETE PROBABILITIES
    subroutine choose_cell (cell, P, N, Pmax)
      implicit none
      integer :: cell, N
      real(8) :: Pmax, r
      real(8), dimension(:), allocatable :: P

!     choose a random index using acceptance rejection
      do

!       choose random cell
        cell = rand_int(N)
        call random_number(r)
        if (r < P(cell)/Pmax) exit
      end do

    end subroutine choose_cell
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE A PARTICLE AND SIGN INDEX UNIFORMLY FROM A GIVEN CELL
    subroutine choose_part (s, i, cell, cref)
      implicit none
      integer :: s, i, cell, opt
      real(8) :: r
      type(Tcref) :: cref

!     choose sign
      s = 0
      call random_number(r)
      if (r < 1.d0*cref%N(1,cell)/(cref%N(0,cell)+cref%N(1,cell))) s = 1

!     choose random particle with appropriate sign
      i = cref%ref(s,cell,rand_int(cref%N(s,cell)))

    end subroutine choose_part
!----------------------------------------------------------------------------------------------------------------------------------

!   TRIAL PARTICLE CREATION STEP
    subroutine create (cell, bounds, cref, f0, floc, part, sim, opt, flag)
      implicit none
      logical :: flag
      integer :: cell, s, opt
      real(8) :: r, a, H, f, x
      real(8), dimension(3) :: b, V, c
      type(Tbounds) :: bounds
      type(Tcref) :: cref
      type(TfMB) :: f0, feq
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tsim) :: sim

!     local equilibrium
      feq%n = floc%n(cell)
      feq%u = floc%u(cell,:)
      feq%c = floc%c(cell)

!     apply bound factors
      a    = cref%a(cell)   * bounds%Ya_coll * bounds%r_coll
      b(1) = cref%b(cell,1) * bounds%Yb_coll * bounds%r_coll
      b(2) = cref%b(cell,2) * bounds%Yb_coll * bounds%r_coll
      b(3) = cref%b(cell,3) * bounds%Yb_coll * bounds%r_coll

!     generate random deviates
      call random_number(r)
      H = r*a
      call random_number(r)
      V(1) = (2*r-1)*b(1)
      call random_number(r)
      V(2) = (2*r-1)*b(2)
      call random_number(r)
      V(3) = (2*r-1)*b(3)

!     velocity
      c = cref%ca(cell) * V/sqrt(H) + cref%ua(cell,:)

!     evaluate distribution
      f = srpi**3 * cref%ca(cell)**3/cref%na(cell) * ( eval_fMB(c,feq) - eval_fMB(c,f0) )

!     particle sign
      s = sign_index(f)

!     acceptance criteria
      if ( ( opt==0 ) .OR. ( opt==2*s-1 ) ) then
      if ( abs(f)**0.4 > H ) then

!       add particle to simulation
        x = uniform_x (cref%x(cell), sim%dx)
        call add_ref (c, x, s, cell, cref, f0, floc, part, sim)

!       update bound factors if necessary
        call update_bounds (H, V, a, b, bounds, sum(part%N))

!       set flag to update bounds for that cell
        flag = .true.
      end if
      end if

    end subroutine create
!----------------------------------------------------------------------------------------------------------------------------------

!   CREATE CELL BASED CROSS REFERENCING
    subroutine cross_ref (Ncell, cref, floc, part)
      implicit none
      integer :: Ncell, i, j, k, s
      type(Tcref) :: cref
      type(Tfloc) :: floc
      type(Tpart) :: part

!     initialize cross referencing
      do k = 0, 1
      do i = 1, Ncell
        cref%N(k,i) = 0
      end do
      end do

!     initialize local properties
      do i = 1, Ncell
        cref%S(i)   = 0
        cref%U(i,:) = (/0.d0,0.d0,0.d0/)
        cref%C(i)   = 0.d0
      end do

!     create cross reference list
      do k = 0, 1
      s = 2*k-1
      do j = 1, part%N(k)
        i = part%cell(k,j)

!       warn if the maximum number of particles per cell is exceeded
        if (cref%N(k,i) >= Ncmax) call stop_exceed_Ncmax (sum(part%N(:)))

!       cross referencing
        cref%N(k,i) = cref%N(k,i) + 1
        cref%ref(k,i,cref%N(k,i)) = j
        cref%idx(k,j) = cref%N(k,i)

!       deviational local properties
        cref%S(i)   = cref%S(i)   + s
        cref%U(i,:) = cref%U(i,:) + s * part%c(k,j,:)
        cref%C(i)   = cref%C(i)   + s * dot_product( part%c(k,j,:), part%c(k,j,:) )
      end do
      end do

!     compute local properties
      do i = 1, Ncell
        floc%n(i)   = 1 + sim%Xi * cref%S(i)
        floc%u(i,:) = sim%Xi * cref%U(i,:) / floc%n(i)
        floc%c(i)   = sqrt( ( 1 + sim%Xi * cref%C(i) / 1.5d0 ) / floc%n(i) - dot_product( floc%u(i,:), floc%u(i,:) ) / 1.5d0 )
      end do

!     initialize conservation cross-referencing
      cref%NG(:) = (/0,0/)
      do k = 0, 1
      do j = 1, part%N(k)
        cref%G(k,j) = .false.
      end do
      end do

    end subroutine cross_ref
!----------------------------------------------------------------------------------------------------------------------------------

!   DELETE RANDOM PARTICLE FROM LIST OF PREVIOUSLY-GENERATED PARTICLES
    subroutine delete (cell, cref, floc, part, sim, opt)
      implicit none
      integer :: cell, s, i, opt
      real(8) :: r
      type(Tcref) :: cref
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tsim) :: sim

!     choose sign at random, if not specified
      if (opt==0) then
        s = 0
        call random_number(r)
        if (r < 1.d0*cref%NG(1)/(cref%NG(0)+cref%NG(1))) s = 1

!     otherwise, assign to specified value
      else
        s = (opt+1)/2
      end if

!     choose index
      i = rand_int(cref%NG(s))
      i = cref%Glist(s,1,i)

!     cell in which particle is to be removed
      cell = part%cell(s,i)

!     remove particle
      call remove_ref (s, i, cref, floc, part, sim)

    end subroutine delete
!----------------------------------------------------------------------------------------------------------------------------------

!   EVALUATE MAXWELLIAN
    function eval_fMB (c, f0)
      implicit none
      real(8) :: eval_fMB
      real(8), dimension(3) :: c
      type(TfMB) :: f0

      eval_fMB = f0%n/(srpi*f0%c)**3 * exp (-dot_product(c-f0%u,c-f0%u)/f0%c**2)

    end function eval_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM PARTICLE GENERATION STEPS UNTIL ONE ADDITIONAL PARTICLE IS GENERATED
    subroutine generate (bounds, cref, f0, floc, part, sim, opt)
      implicit none
      logical :: flag
      integer :: opt, N0
      type(Tbounds) :: bounds
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tsim) :: sim

!     initial number of particles
      N0 = sum(part%N)

!     repeat until single particle is generated
      do

!       choose cell based on index distribution

        call choose_cell (cell, cref%J, sim%Ncell, cref%Jmax)

!       trial particle creation step
        call create (cell, bounds, cref, f0, floc, part, sim, opt, flag)

!       exit criterion
        if ( sum(part%N) == N0 + 1 ) exit
      end do

!     update ratio-of-uniforms bounds
      call change_collision_bound (cell, bounds, cref, f0, floc, sim)

    end subroutine generate
!----------------------------------------------------------------------------------------------------------------------------------

!   VECTOR MAGNINTUDE
    function magnitude (c)
      implicit none
      real(8) :: magnitude
      real(8), dimension(3) :: c

      magnitude = sqrt(dot_product(c,c))

    end function magnitude
!----------------------------------------------------------------------------------------------------------------------------------

!   CORRECT TOTAL SIGN PARITY
    subroutine parity (bounds, cref, f0, floc, part, sim)
      implicit none
      integer :: cell
      real(8) :: r
      type(Tbounds) :: bounds
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tsim) :: sim

!     for odd parity, change it to even (if possible)
      if ((mod(abs(cref%DS),2)==1).AND.(sum(cref%NG(:))>0)) then

!       with equal probability increase/decrease generated particle population
        call random_number(r)
        if (r<0.5d0) then

!         delete previously generated particle
          call delete (cell, cref, floc, part, sim, 0)

!         update ratio-of-uniforms bounds
          call change_collision_bound (cell, bounds, cref, f0, floc, sim)
        else

!         otherwise, generate additional particle
          call generate (bounds, cref, f0, floc, part, sim, 0)
        end if
      end if

    end subroutine parity
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A RANDOM INTEGER FROM 1 TO N
    function rand_int(N)
      implicit none
      integer :: N, rand_int
      real(8) :: r

      call random_number(r)
      rand_int = int(r*N) + 1

    end function rand_int
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM REGENERATE PREVIOUSLY GENERATED PARTICLES WITH OPPOSITE SIGN UNTIL OPTIMAL SIGN DEFICIT IS OBTAINED
    subroutine regenerate (bounds, cref, f0, floc, part, sim)
      implicit none
      integer :: s, DS, i, cell
      real(8) :: r
      type(Tbounds) :: bounds
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tsim) :: sim

!     perform this step only if the mass deficit is nonzero and there are generated particles to resample
      if ((cref%DS/=0).AND.(sum(cref%NG(:))>0)) then

!       determine optimal sign deficit
        s = sign(1,cref%DS)
        if (s==-1) then
          DS = min(0,cref%DS+2*cref%NG(0))
        else
          DS = max(0,cref%DS-2*cref%NG(1))
        end if

!       perform regeneration step until optimal sign deficit is obtained
        do

!         check sign deficit, and exit (if appropriate)
          if (cref%DS==DS) exit

!         choose cell based on index distribution
          call choose_cell (cell, cref%J, sim%Ncell, cref%Jmax)

!         delete/generate particle in random order
          call random_number(r)
          if (r<0.5D0) then

!           delete first, then generate
            call delete (cell, cref, floc, part, sim, s)
            call change_collision_bound (cell, bounds, cref, f0, floc, sim)
            call generate (bounds, cref, f0, floc, part, sim, -s)
          else

!           generate first, then delete
            call generate (bounds, cref, f0, floc, part, sim, -s)
            call delete (cell, cref, floc, part, sim, s)
            call change_collision_bound (cell, bounds, cref, f0, floc, sim)
          end if
        end do
      end if

    end subroutine regenerate
!----------------------------------------------------------------------------------------------------------------------------------

!   REMOVE A PARTICLE FROM SIMULATION MAINTAINING CROSS REFERENCING
    subroutine remove_ref (s, i, cref, floc, part, sim)
      implicit none
      integer :: i, j, k, s, ss
      real(8) :: Tloc
      real(8), dimension(3) :: c
      type(Tcref) :: cref
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Tsim) :: sim

!     particle sign and velocity
      ss = 2*s-1
      c = part%c(s,i,:)

!     update cross referencing for deleted particle
      j = part%cell(s,i)
      k = cref%ref(s,j,cref%N(s,j))
      cref%ref(s,j,cref%idx(s,i)) = k
      if (i/=part%N(s)) cref%ref(s,part%cell(s,part%N(s)),cref%idx(s,part%N(s))) = i
      cref%idx(s,k) = cref%idx(s,i)
      if (i/=part%N(s)) cref%idx(s,i) = cref%idx(s,part%N(s))
      cref%N(s,j) = cref%N(s,j) - 1

!     reassign highest numbered particle in the simulation
      if (i /= part%N(s)) then
        part%x(s,i) = part%x(s,part%N(s))
        part%c(s,i,:) = part%c(s,part%N(s),:)
        part%cell(s,i) = part%cell(s,part%N(s))
        part%del(s,i) = .false.
        cref%idx(s,i) = cref%idx(s,part%N(s))
        cref%ref(s,part%cell(s,i), cref%idx(s,i)) = i
      end if

!     update conservation enforcement tracking
      cref%DS = cref%DS - ss

!     update local properties in affected cell
      cref%S(cell)   = cref%S(cell) - ss
      cref%U(cell,:) = cref%U(cell,:) - ss*c
      cref%C(cell)   = cref%C(cell) - ss*dot_product(c,c)
      floc%n(cell)   = 1 + sim%Xi * cref%S(cell)
      floc%u(cell,:) = sim%Xi * cref%U(cell,:) / floc%n(cell)
      Tloc           = ( 1 + sim%Xi * cref%C(cell) / 1.5d0 ) / floc%n(cell) - dot_product( floc%u(cell,:), floc%u(cell,:) ) / 1.5d0

!     determine local most probable molecular speed if temperature is positive
      if ( Tloc <= 0.d0 ) then

!       stop simulation if a negative temperature is encountered
        call stop_Tneg (sum(part%N))
      else

!       update most probable molecular speed
        floc%c(cell)   = sqrt( Tloc )
      end if

!     update cell index distribution
      cref%Jsum = cref%Jsum - cref%J(cell)
      cref%J(cell) = floc%n(cell) * ( cref%F(cell) + sum(cref%N(:,cell)) )
      cref%Jsum = cref%Jsum + cref%J(cell)
      if ( cref%J(cell) > cref%Jmax ) cref%Jmax = cref%J(cell)

!     correct list of generated particles if affected
      if ( cref%G(s,i) .OR. cref%G(s,part%N(s)) ) then

!       if both deleted and highest-numbered particle were in list of generated particles
        if ( cref%G(s,i) .AND. cref%G(s,part%N(s)) ) then
          if ( part%N(s) /= cref%Glist(s,1,cref%NG(s)) ) then
            cref%Glist(s,1,cref%Gidx(s,part%N(s))) = i
            cref%Glist(s,1,cref%Gidx(s,i)) = cref%Glist(s,1,cref%NG(s))
            cref%Gidx(s,cref%Glist(s,1,cref%NG(s))) = cref%Gidx(s,i)
            cref%Gidx(s,i) = cref%Gidx(s,part%N(s))
          end if
          cref%NG(s) = cref%NG(s) - 1

!       if only the deleted particle was in list of generated particles
        else if ( cref%G(s,i) ) then
          cref%Glist(s,1,cref%Gidx(s,i)) = cref%Glist(s,1,cref%NG(s))
          cref%Gidx(s,cref%Glist(s,1,cref%NG(s))) = cref%Gidx(s,i)
          cref%NG(s) = cref%NG(s) - 1
          cref%G(s,i) = .false.

!       if only the highest-numbered particle was in the list of generated particles
        else
          cref%Glist(s,1,cref%Gidx(s,part%N(s))) = i
          cref%Gidx(s,i) = cref%Gidx(s,part%N(s))
          cref%G(s,i) = .true.
        end if
      end if

!     decreased number of particles in the simulation
      part%N(s) = part%N(s) - 1

    end subroutine remove_ref
!----------------------------------------------------------------------------------------------------------------------------------

!   SET COLLISION GENERATION BOUNDS BASED ON LINEARIZED CONDITIONS
    subroutine set_collision_bounds (bounds, cref, f0, floc, sim)
      use numerical_constants
      use types
      use sampling_bound_constants
      implicit none
      integer :: i
      real(8) :: na, dn, ca, dc
      real(8), dimension(3) :: ua, du
      type(Tbounds) :: bounds
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tfloc) :: floc
      type(Tsim) :: sim

      do i = 1, sim%Ncell

!       compute average properties for generating particles at the boundaries
        na = (f0%n + floc%n(i))/2
        ua = (f0%u + floc%u(i,:))/2
        ca = (f0%c + floc%c(i))/2

!       compute the absolute property differences for estimating the sampling bounds
        dn = (floc%n(i) - f0%n) / na
        du = (floc%u(i,:) -f0%u) / ca
        dc = (floc%c(i) - f0%c) / ca

!       compute linearized sampling bounds
        cref%a(i) = ( abs(dn-3.d0*dc)*a0 + 2.d0*abs(du(1))*an + 2.d0*abs(du(2))*an + 2.d0*abs(du(3))*an + 2.d0*abs(dc)*a2 )**0.4
        cref%b(i,1) = ( abs(dn-3.d0*dc)*b0 + 2.d0*abs(du(1))*bn_n + 2.d0*abs(du(2))*bn_s + 2.d0*abs(du(3))*bn_s                   &
                                                                                                           + 2.d0*abs(dc)*b2 )**0.2
        cref%b(i,2) = ( abs(dn-3.d0*dc)*b0 + 2.d0*abs(du(1))*bn_s + 2.d0*abs(du(2))*bn_n + 2.d0*abs(du(3))*bn_s                   &
                                                                                                           + 2.d0*abs(dc)*b2 )**0.2
        cref%b(i,3) = ( abs(dn-3.d0*dc)*b0 + 2.d0*abs(du(1))*bn_s + 2.d0*abs(du(2))*bn_s + 2.d0*abs(du(3))*bn_n                   &
                                                                                                           + 2.d0*abs(dc)*b2 )**0.2

!       integral bound on difference between local and global equilibrium state
        cref%F(i) = 20.d0/srpi**3 * na/sim%Xi * cref%a(i)*product(cref%b(i,:)) * bounds%Ya_coll*bounds%Yb_coll**3*bounds%r_coll**4

!       store average properties
        cref%na(i)   = na
        cref%ua(i,:) = ua
        cref%ca(i)   = ca
      end do

!     initialize index cell distribution
      cref%Jmax = 0.d0
      cref%Jsum = 0.d0

!     compute cell index distribution
      do i = 1, sim%Ncell
        cref%J(i) = floc%n(i) * ( cref%F(i) + sum(cref%N(:,i)) )
        cref%Jsum = cref%Jsum + cref%J(i)
        if ( cref%J(i) > cref%Jmax ) cref%Jmax = cref%J(i)
      end do

    end subroutine set_collision_bounds
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD EXPONENTIALLY DISTRIBUTED TIME STEP TO THE CURRENT COLLISION TIME
    subroutine step (first, coll, cref, floc, part, time, sim)
      implicit none
      logical :: coll, first
      real(8) :: Kn, r, a
      type(Tcref) :: cref
      type(Tfloc) :: floc
      type(Tpart) :: part
      type(Ttime) :: time
      type(Tsim) :: sim

!     check to see if the simulation is empty
      if (sum(part%N) < 1) then

!       advance cell a full time step if it is empty
        time%tc = time%t
        coll = .true.

!     normal routine
      else

!       create new time step only if one is not available from previous advection step
        if ( (.NOT.first) .OR. (time%dtc < 0.d0) ) then

!         time step parameter
          a = sim%Kn*cref%Jsum 

!         generate time step
          call random_number(r)
          time%dtc = -log(r)/a
        end if

!       advance collision time only if it does not exceed advection time
        if (time%tc + time%dtc < time%t) then
          time%tc = time%tc + time%dtc
        else
          coll = .true.
        end if
      end if

    end subroutine step
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0, '(a55)') '-------------------------------------------------------'
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES PER CELL WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Ncmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a50)') '[',tcpu,Np,']: ERROR: Ncmax exceeded--increase Ncmax parameter'
      write (0, '(a55)') '-------------------------------------------------------'
      flush (0)
      stop

    end subroutine stop_exceed_Ncmax
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_Tneg (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a72)') '------------------------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a53)') '[',tcpu,Np,']: ERROR: negative temperature--increase Xi parameter'
      write (0, '(a72)') '------------------------------------------------------------------------'
      flush (0)
      stop

    end subroutine stop_Tneg
!----------------------------------------------------------------------------------------------------------------------------------

!   GENERATE UNIFORMLY DISTRIBUTED SPATIAL POSITION WITHIN A CELL
    function uniform_x (x0, dx)
      implicit none
      real(8) :: r, x0, dx, uniform_x

      call random_number(r)
      uniform_x = x0 + r*dx

    end function uniform_x
!----------------------------------------------------------------------------------------------------------------------------------

!   UPDATE BOUND FACTORS IF NECESSARY
    subroutine update_bounds (H, V, a, b, bounds, Np)
      implicit none
      logical :: change
      integer :: Np, disp, dir
      real(8) :: H, a, tcpu, Vmax
      real(8), dimension(3) :: V, b
      type(Tbounds) :: bounds

      change = .false.

!     test bound on dependent variable
      if (H > a/bounds%r_coll) then
        bounds%Ya_coll = (H/a) * bounds%Ya_coll * bounds%r_coll
        change = .true.
      end if

!     test bound on independent variable
      Vmax = max( abs(V(1)/b(1)), abs(V(2)/b(2)), abs(V(3)/b(3)) )
      if (Vmax > 1.d0/bounds%r_coll) then
        bounds%Yb_coll = Vmax * bounds%Yb_coll * bounds%r_coll
        change = .true.
      end if

!     warn user if bound factors were updated
      if (change) then
        call cpu_time(tcpu)
        write (0, '(a55)') '-------------------------------------------------------'
        write (0, '(a1,es9.3,i9,a36)') '[',tcpu,Np,']: ATTENTION: bound factors changed!'
        write (0, '(a55)') '-------------------------------------------------------'
        flush (0)
      end if

    end subroutine update_bounds
!----------------------------------------------------------------------------------------------------------------------------------

  end subroutine collision


!==================================================================================================================================
! EFFECTIVE BODY FORCE ROUTINE
!==================================================================================================================================

! GENERATE PARTICLES TO ACCOMODATE VARIOUS KINDS OF EFFECTIVE BODY FORCE TERMS
  subroutine force (f0, part, sim)
    use types
    use numerical_constants
    implicit none
    integer :: j
    integer, dimension(0:1) :: NG
    type(TfMB) :: f0
    type(Tpart) :: part
    type(Tsim) :: sim

!   particle generation routine with mass conservation
    call generate_force (0.5d0, f0, part, sim, NG, .false.)
    j = NG(0)
    NG(0) = NG(1)
    NG(1) = j
    call generate_force (0.5d0, f0, part, sim, NG, .true.)

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD PARTICLE TO SIMULATION
    subroutine add (c, x, s, cell, part)
      implicit none
      integer :: s, cell
      real(8) :: x
      real(8), dimension(3) :: c
      type(Tpart) :: part

!     warn if the maximum number of particles is exceeded
      if (maxval(part%N)>=Nmax) call stop_exceed_Nmax (sum(part%N))

!     add new velocity sample
      part%N(s) = part%N(s) + 1
      part%c(s, part%N(s),:) = c
      part%x(s, part%N(s)) = x
      part%cell(s, part%N(s)) = cell
      part%del(s, part%N(s)) = .false.

    end subroutine add
!----------------------------------------------------------------------------------------------------------------------------------

!   GENERATE PARTICLES FOR EFFECTIVE BODY FORCE TERMS
    subroutine generate_force (frac, f0, part, sim, NG, opt)
      implicit none
      logical :: esc, opt, sgn
      integer :: N, p, s, cell
      integer, dimension(0:1) :: NG
      real(8) :: r, H, f, c2, frac, x
      real(8), dimension(3) :: V, c
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tsim) :: sim

!     use number of trial particles as exit condition when opt=.false.
      if (.NOT.opt) then

!       determine the number of trial particles
        N = int(sim%Nf*frac)
        call random_number(r)
        if (r < sim%Nf*frac-N) N = N + 1

!       initialize the number of generated particles
        NG = (/0,0/)
      else

!       invert the sign of the desired number of particles to generate
        NG = -NG
      end if

!     generate particles
      p = 0
      do

!       exit criterion
        if (opt) then
          if (sum(NG)==0) exit
        else
          if (p==N) exit
        end if

!       advance do loop counter
        p = p + 1

!       generate random deviates
        call random_number(r)
        H = r*sim%af
        call random_number(r)
        V(1) = (2*r-1)*sim%bf(1)
        call random_number(r)
        V(2) = (2*r-1)*sim%bf(2)
        call random_number(r)
        V(3) = (2*r-1)*sim%bf(3)

!       velocity
        c = V/sqrt(H)

!       evaluate distribution
        c2 = dot_product(c,c)
        f =  ( sim%Pz*c(3) + sim%Tz*c(3)*(2.5d0-c2) ) * exp(-c2)

!       particle sign
        s = sign_index(f)

!       acceptance criteria
        sgn = NG(s) /= 0
        if (abs(f)**0.4 > H) then
        if ( (.NOT.opt) .OR. sgn ) then

!         adjust particle velocity based on reference distribution
          c = f0%c*c + f0%u

!         particle position
          call random_number(r)
          x = r*sim%Lx

!         determine cell number
          call locate (esc, cell, x, sim)

!         add particle to simulation
          call add (c, x, s, cell, part)
          NG(s) = NG(s) + 1
        end if
        end if
      end do

    end subroutine generate_force
!----------------------------------------------------------------------------------------------------------------------------------

!   LOCATE A PARTICLE TO A PARTICULAR CELL OR BOUNDARY SURFACE
    subroutine locate (esc, cell, x, sim)
      use types
      implicit none
      logical :: esc
      integer :: cell
      real(8) :: x
      type(Tsim) :: sim

!     calculate cell number
      cell = 1 + floor( sim%Ncell * x / sim%Lx )

!     determine boundary cell, if particle escapes
      if ( cell < 1 ) then
        esc = .true.
        cell = 1
      else if ( cell > sim%Ncell ) then
        esc = .true.
        cell = 2
      else
        esc = .false.
      end if

    end subroutine locate

!----------------------------------------------------------------------------------------------------------------------------------
!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0, '(a55)') '-------------------------------------------------------'
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

  end subroutine force

!==================================================================================================================================
